#include <sys/param.h>
#include <sys/systm.h>
#include <sys/conf.h>
#include <sys/exec.h>
#include <sys/lkm.h>
#include <sys/device.h>
#include <sys/ucred.h>
#include <sys/proc.h>

#include <dev/pci/pcivar.h>
#include <dev/pci/pciidereg.h>
#include <dev/pci/pciidevar.h>

#include <dev/isa/isavar.h>
#include <dev/isa/fdreg.h>
#include <dev/isa/fdcvar.h>

#include <dev/ic/mc146818reg.h>
#include <i386/isa/nvram.h>

#include <sys/mount.h>
#include <sys/syscallargs.h>

#include <lib/libsa/stand.h>
#include "hotswap.h"

extern struct cfdriver pciide_cd,fdc_cd,fd_cd;
extern struct cfattach fd_ca;

int fdprint(void *,const char *);
void *fd_nvtotype (char *, int, int);

static int fd_activate(struct device *,enum devact);
static int (*pre_activate)(struct device *,enum devact);

static u_int32_t	myfd;

static int hotswap_mod(struct lkm_table *t,int cmd)
{
	switch (cmd) {
		case LKM_E_LOAD:
			pre_activate = fd_ca.ca_activate;
			fd_ca.ca_activate = fd_activate;
			break;
		case LKM_E_UNLOAD:
			fd_ca.ca_activate = pre_activate;
	}
	return 0;
}

static int fd_activate(struct device *dev,enum devact op)
{
	u_int32_t	mask = 1 << dev->dv_unit;

	if (!(mask & myfd)) return EINVAL;
	return 0;
}

static int atapi_doit(int op,int ide,int channel)
{
	struct pciide_softc	*ide_sc;
	struct channel_softc	*chp;
	int	err,s;

	if (ide >= pciide_cd.cd_ndevs) return EINVAL;
	ide_sc = pciide_cd.cd_devs[ide];
	if (!ide_sc) return EINVAL;
	chp = &ide_sc->pciide_channels[channel].wdc_channel;
	if (!chp) return EINVAL;

	if (op) {
		wdcattach(chp);
		ide_sc->sc_wdcdev.set_modes(chp);
		return 0;
	}
	else if (chp->atapibus) {
		s = splbio();
		wdc_kill_pending(chp);
		splx(s);
		if ((err = config_detach(chp->atapibus,0)))
			printf("atapi detach: %s\n",strerror(err));
		else chp->atapibus = NULL;
		return err;
	}
	else return EINVAL;
}

				
static int fd_doit(int op,int fdc,int drive)
{
	struct fdc_softc	*fdcp;
/* XXX stolen from dev/isa/fd.c */
	struct fdc_attach_args {
		int	drive;
		void	*type;
	} fa;
	struct device	*fd_dev;
	u_int32_t	mask;
	int		type,i,err;

	if (fdc >= fdc_cd.cd_ndevs) return EINVAL;
	fdcp = fdc_cd.cd_devs[fdc];
	if (!fdcp) return EINVAL;

	if (op) {
		fa.drive = drive;
		config_search(NULL,&fdcp->sc_dev,&fa);
		type = mc146818_read(NULL, NVRAM_DISKETTE);
		fa.type = fd_nvtotype(fdcp->sc_dev.dv_xname,type,fa.drive);
			
		if ((fd_dev = config_found(&fdcp->sc_dev,(void *)&fa,fdprint))) {
			mask = 1 << fd_dev->dv_unit;
			myfd |= mask;
			return 0;
		}
		else return ENOENT;
	}
	else {
		for (i=0; i<fd_cd.cd_ndevs; i++) {
			struct device	*dev = (struct device *) fd_cd.cd_devs[i];
			if (dev && dev->dv_parent == (struct device *) fdcp
				&& (struct device *) fdcp->sc_fd[drive] == dev)
			
			{
				char	buf[100];
				strcpy(buf,dev->dv_xname);

/* XXX: returns EOPNOTSUPP normally (missing detach routine) but it looks OK */
				if ((err = config_detach(dev,0)) && err != EOPNOTSUPP) {
					printf("detach %s: %s\n",buf,
							strerror(err));
				}
				fd_cd.cd_devs[i] = NULL;
				printf("%s detached\n",buf);
			}
		}
		return 0;
	}
}

static int hotswap_call(struct proc *p,void *v,register_t *retval)
{
	struct sys_hotswap_args	*uap = v;
	int	op = SCARG(uap,op),
		ctrl = SCARG(uap,ctrl),
		unit = SCARG(uap,unit);

	if (p->p_cred->p_ruid && p->p_cred->pc_ucred->cr_uid) return EPERM;
		
	switch (op) {
		case HOTSWAP_IN_FD: 	return fd_doit(1,ctrl,unit);
		case HOTSWAP_IN_CD: 	return atapi_doit(1,ctrl,unit);
		case HOTSWAP_OUT_FD: 	return fd_doit(0,ctrl,unit);
		case HOTSWAP_OUT_CD: 	return atapi_doit(0,ctrl,unit);
		default:		return EINVAL;
	}
}



static struct sysent hotswap_ent = {
#if __NetBSD_Version__ >= 106000000
	2,sizeof(struct sys_hotswap_args),0,hotswap_call
#else
	2,sizeof(struct sys_hotswap_args),hotswap_call
#endif
};

MOD_SYSCALL("hotswap",-1,&hotswap_ent);

int hotswap(struct lkm_table *t,int cmd,int ver)
{
	DISPATCH(t,cmd,ver,hotswap_mod,hotswap_mod,lkm_nofunc)
}
